from datetime import datetime, timedelta
from unittest.mock import AsyncMock, Mock

try:
    from builtins import aiter, anext
except ImportError:
    from typing import AsyncIterable, AsyncIterator, Awaitable, TypeVar

    _T = TypeVar("_T")

    def aiter(iterable: AsyncIterable[_T]) -> AsyncIterator[_T]:
        """Return an AsyncIterator for an AsyncIterable object."""
        return iterable.__aiter__()

    def anext(iterator: AsyncIterator[_T]) -> Awaitable[_T]:
        """Return the next item from the async iterator."""
        return iterator.__anext__()


import pytest

from combined_energy import helpers


class TestReadingsIterator:
    def test_next_range_start__where_no_request_made(self):
        target = helpers.ReadingsIterator(
            AsyncMock(helpers.CombinedEnergy), increment=10
        )

        actual = target.next_range_start

        assert actual is None

    def test_next_range_start__where_successful_request_made(self):
        target = helpers.ReadingsIterator(
            AsyncMock(helpers.CombinedEnergy), increment=10
        )
        target._last_end = datetime(2022, 2, 22, 22, 22, 22)

        actual = target.next_range_start

        assert actual == datetime(2022, 2, 22, 22, 22, 22)

    def test_next_range_start__where_timedelta_set(self, monkeypatch):
        monkeypatch.setattr(helpers, "now", lambda: datetime(2022, 2, 22, 22, 22, 22))

        target = helpers.ReadingsIterator(
            AsyncMock(helpers.CombinedEnergy),
            increment=10,
            initial_delta=timedelta(minutes=22),
        )

        actual = target.next_range_start

        assert actual == datetime(2022, 2, 22, 22, 00, 22)

    @pytest.mark.asyncio
    async def test_check_for_log_session_restart__check_behaviour(self):
        mock_client = AsyncMock(helpers.CombinedEnergy)
        target = helpers.ReadingsIterator(
            mock_client, increment=10, log_session_reset_count=2
        )

        await target._check_for_log_session_restart()
        mock_client.start_log_session.assert_not_called()

        target._empty.append(False)
        target._empty.append(True)

        await target._check_for_log_session_restart()
        mock_client.start_log_session.assert_not_called()

        target._empty.append(True)

        await target._check_for_log_session_restart()
        mock_client.start_log_session.assert_called()

    @pytest.mark.asyncio
    async def test_iterator(self):
        mock_client = AsyncMock(
            helpers.CombinedEnergy,
            readings=AsyncMock(
                side_effect=[
                    Mock(range_count=1, range_end=datetime(2022, 2, 22, 22, 00, 21)),
                    Mock(range_count=1, range_end=datetime(2022, 2, 22, 22, 00, 22)),
                    Mock(range_count=0, range_end=datetime(2022, 2, 22, 22, 00, 22)),
                    Mock(range_count=0, range_end=datetime(2022, 2, 22, 22, 00, 22)),
                    Mock(range_count=3, range_end=datetime(2022, 2, 22, 22, 00, 25)),
                ]
            ),
        )
        target = helpers.ReadingsIterator(
            mock_client, increment=10, log_session_reset_count=2
        )
        iterator = aiter(target)

        # Check first request
        await anext(iterator)
        assert target.next_range_start == datetime(2022, 2, 22, 22, 00, 21)
        assert mock_client.start_log_session.call_count == 1

        # Check second request
        await anext(iterator)
        assert target.next_range_start == datetime(2022, 2, 22, 22, 00, 22)
        assert mock_client.start_log_session.call_count == 1

        # Check third request
        await anext(iterator)
        assert target.next_range_start == datetime(2022, 2, 22, 22, 00, 22)
        assert mock_client.start_log_session.call_count == 1

        # Check forth request
        await anext(iterator)
        assert target.next_range_start == datetime(2022, 2, 22, 22, 00, 22)
        assert mock_client.start_log_session.call_count == 1

        # Check fifth request
        await anext(iterator)
        assert target.next_range_start == datetime(2022, 2, 22, 22, 00, 25)
        assert mock_client.start_log_session.call_count == 2



"""Expected exception types."""


class CombinedEnergyError(Exception):
    """Common Error for Combined Energy API."""


class CombinedEnergyTimeoutError(CombinedEnergyError):
    """Timeout occurred accessing API."""


class CombinedEnergyAuthError(CombinedEnergyError):
    """Error occurred with Authentication."""


class CombinedEnergyPermissionError(CombinedEnergyError):
    """Don't have permission to access particular item."""



"""API Client for Combined Energy API."""
from . import constants, exceptions
from .client import CombinedEnergy
from .helpers import ReadingsIterator

__all__ = ("constants", "exceptions", "CombinedEnergy", "ReadingsIterator")



from pathlib import Path
import sys

import pytest

HERE = Path(__file__).parent

# Configure src location
SRC = (HERE.parent / "src").absolute()
sys.path.insert(0, SRC.as_posix())


@pytest.fixture
def api_responses() -> Path:
    return HERE / "fixtures" / "api-responses"



import pytest

from combined_energy import utils


def test_energy_to_power():
    actual = utils.energy_to_power(5.0, 5.0)

    assert actual == 3.6


@pytest.mark.parametrize("seconds", (0.0, -1.0))
def test_energy_to_power__raises_value_error_for_invalid_seconds(seconds):
    with pytest.raises(ValueError, match=r"Seconds must be a positive float\."):
        utils.energy_to_power(5.0, seconds)



"""Helpers for using the API client."""
from collections import deque
from datetime import datetime, timedelta
import logging
from typing import AsyncIterator, Optional

from .client import CombinedEnergy
from .models import Readings

LOGGER = logging.getLogger(__package__)

now = datetime.now


class ReadingsIterator:
    """Iterator that returns readings with a certain interval specified."""

    __slots__ = ("client", "increment", "initial_delta", "_last_end", "_empty")

    def __init__(
        self,
        client: CombinedEnergy,
        *,
        increment: int,
        initial_delta: timedelta = None,
        log_session_reset_count: int = 3,
    ):
        """Initialise iterator."""
        self.client = client
        self.increment = increment
        self.initial_delta = initial_delta

        self._last_end: Optional[datetime] = None
        # Prefill with False
        self._empty = deque([False] * log_session_reset_count, log_session_reset_count)

    @property
    def next_range_start(self) -> Optional[datetime]:
        """Calculate the next range start value."""
        if self._last_end:
            return self._last_end
        if self.initial_delta:
            return now() - self.initial_delta

    async def _check_for_log_session_restart(self) -> None:
        """
        Check if a start log session message is required.

        This is a strange behaviour of this API that requires the client to
        call start new log session. In the dashboard app this is covered up
        with a dialog that pops up periodically.
        """
        # Reset if deque is full of True statuses
        if all(self._empty):
            LOGGER.info("Log session expired, restarting...")
            await self.client.start_log_session()

    async def __anext__(self) -> Readings:
        """Async Next implementation."""
        await self._check_for_log_session_restart()

        readings = await self.client.readings(
            self.next_range_start, None, self.increment
        )

        # Update state
        self._empty.append(readings.range_count == 0)
        self._last_end = readings.range_end

        return readings

    async def __aiter__(self) -> AsyncIterator[Readings]:
        """Async iterator implementation."""

        # Start log session up front if required
        await self.client.start_log_session()

        while True:
            yield await self.__anext__()



"""API Schema model."""
from datetime import datetime, timedelta
from typing import Any, Callable, Literal

from pydantic import BaseModel, Field, ValidationError

from .constants import LOGGER, DeviceType
from .utils import energy_to_power

now = datetime.now
OptionalFloatList = list[None | float]
OptionalStrList = list[None | str]


class Login(BaseModel):
    """Response from Login."""

    status: str
    expire_minutes: int = Field(alias="expireMins")
    jwt: str

    # Capture time login was created
    created: datetime = Field(default_factory=now)

    def expires(self, expiry_window: int) -> datetime:
        """Calculate when this login expires."""
        offset = timedelta(minutes=self.expire_minutes, seconds=-expiry_window)
        return self.created + offset


class _CommonModel(BaseModel):
    status: str
    installation_id: int = Field(alias="installationId")


class User(BaseModel):
    """Individual user."""

    type: str
    id: int
    email: str
    mobile: str
    fullname: str
    dsa_ok: bool = Field(alias="dsaOk")
    show_introduction: None | str = Field(alias="showIntroduction")


class CurrentUser(BaseModel):
    """Current logged in user."""

    status: str
    user: User


class ConnectionStatus(_CommonModel):
    """Connection Status of the monitor."""

    connected: bool
    since: datetime


class ConnectionHistoryEntry(BaseModel):
    """Entry in connection history."""

    connected: bool
    timestamp: datetime = Field(alias="t")
    device: str = Field(alias="d")
    s: float


class ConnectionHistoryRoute(BaseModel):
    """Route in connection history."""

    timestamp: datetime = Field(alias="t")
    device: str = Field(alias="d")


class ConnectionHistory(_CommonModel):
    """Connection History of the monitor."""

    history: list[ConnectionHistoryEntry]
    route: ConnectionHistoryRoute


class Device(BaseModel):
    """Details of a device."""

    device_id: int = Field(alias="deviceId")
    ref_name: str = Field(alias="refName")
    display_name: str = Field(alias="displayName")
    device_type: str = Field(alias="deviceType")
    device_manufacturer: None | str = Field(alias="deviceManufacturer")
    device_model_name: None | str = Field(alias="deviceModelName")
    supplier_device: bool = Field(alias="supplierDevice")
    storage_device: bool = Field(alias="storageDevice")
    consumer_device: bool = Field(alias="consumerDevice")
    status: str
    max_power_supply: None | int = Field(alias="maxPowerSupply")
    max_power_consumption: None | int = Field(alias="maxPowerConsumption")
    icon_override: None | str = Field(alias="iconOverride")
    order_override: None | int = Field(alias="orderOverride")
    category: str


class Installation(_CommonModel):
    """Details of an installation."""

    source: str
    role: str
    read_only: bool = Field(alias="readOnly")
    dmg_id: int = Field(alias="dmgId")
    tags: list[str]

    mqtt_account_kura: str = Field(alias="mqttAccountKura")
    mqtt_broker_ems: str = Field(alias="mqttBrokerEms")

    timezone: str
    street_address: str = Field(alias="streetAddress")
    locality: str
    state: str
    postcode: str

    review_status: str = Field(alias="reviewStatus")
    nmi: str
    phase: int
    org_id: int = Field(alias="orgId")
    brand: str

    tariff_plan_id: int = Field(alias="tariffPlanId")
    tariff_plan_accepted: int = Field(alias="tariffPlanAccepted")

    devices: list[Device]
    pm: dict[str, list[dict]]


class Customer(BaseModel):
    """Individual customer."""

    customer_id: int = Field(alias="customerId")
    phone: None | str
    email: str
    name: str
    primary: bool


class InstallationCustomers(_CommonModel):
    """Response from customers."""

    customers: list[Customer]


class DeviceReadings(BaseModel):
    """Readings for a particular device."""

    device_id: int | None = Field(alias="deviceId", default=None)
    range_start: datetime | None = Field(alias="rangeStart")
    range_end: datetime | None = Field(alias="rangeEnd")
    timestamp: list[datetime]
    sample_seconds: None | list[int] = Field(alias="sampleSecs")


def _device_energy_sample_to_power(
    attribute: str, *, sample_index: int = -1
) -> Callable[[DeviceReadings], None | float]:
    def _energy_sample_as_power(self: DeviceReadings) -> None | float:
        """Generate a power instantaneous power figure in kW."""
        if self.sample_seconds:
            energy_samples = getattr(self, attribute)
            return energy_to_power(
                energy_samples[sample_index],
                self.sample_seconds[sample_index],
            )

    return _energy_sample_as_power


class DeviceReadingsCombiner(DeviceReadings):
    """Readings for the Combiner device."""

    device_type: Literal["COMBINER"] = Field(alias="deviceType")

    energy_supplied: None | list[None | float] = Field(alias="energySupplied")
    energy_supplied_solar: None | list[None | float] = Field(
        alias="energySuppliedSolar"
    )
    energy_supplied_battery: None | list[None | float] = Field(
        alias="energySuppliedBattery"
    )
    energy_supplied_grid: None | list[None | float] = Field(alias="energySuppliedGrid")
    energy_consumed_other: None | list[None | float] = Field(
        alias="energyConsumedOther"
    )
    energy_consumed_other_solar: None | list[None | float] = Field(
        alias="energyConsumedOtherSolar"
    )
    energy_consumed_other_battery: None | list[None | float] = Field(
        alias="energyConsumedOtherBattery"
    )
    energy_consumed_other_grid: None | list[None | float] = Field(
        alias="energyConsumedOtherGrid"
    )
    energy_consumed: None | list[None | float] = Field(alias="energyConsumed")
    energy_consumed_solar: None | list[None | float] = Field(
        alias="energyConsumedSolar"
    )
    energy_consumed_battery: None | list[None | float] = Field(
        alias="energyConsumedBattery"
    )
    energy_consumed_grid: None | list[None | float] = Field(alias="energyConsumedGrid")
    energy_correction: None | list[None | float] = Field(alias="energyCorrection")
    temperature: None | list[None | float]


class DeviceReadingsSolarPV(DeviceReadings):
    """Readings for the Solar PV device."""

    device_type: Literal["SOLAR_PV"] = Field(alias="deviceType")
    operation_status: None | list[None | str] = Field(alias="operationStatus")
    operation_message: None | list[None | str] = Field(alias="operationMessage")

    energy_supplied: None | list[None | float] = Field(alias="energySupplied")

    # Calculate power figures
    power_supply = property(_device_energy_sample_to_power("energy_supplied"))

    def __str__(self) -> str:
        """Convert self to a string representation."""
        if self.sample_seconds:
            return f"Generation: {self.power_supply:0.02f}kW"
        else:
            return "_.__kW"


class DeviceReadingsGridMeter(DeviceReadings):
    """Readings for the Grid Meter device."""

    device_type: Literal["GRID_METER"] = Field(alias="deviceType")
    operation_status: None | list[None | str] = Field(alias="operationStatus")
    operation_message: None | list[None | str] = Field(alias="operationMessage")

    energy_supplied: None | list[None | float] = Field(alias="energySupplied")
    energy_consumed: None | list[None | float] = Field(alias="energyConsumed")
    energy_consumed_solar: None | list[None | float] = Field(
        alias="energyConsumedSolar"
    )
    energy_consumed_battery: None | list[None | float] = Field(
        alias="energyConsumedBattery"
    )
    power_factor_a: None | list[None | float] = Field(alias="powerFactorA")
    power_factor_b: None | list[None | float] = Field(alias="powerFactorB")
    power_factor_c: None | list[None | float] = Field(alias="powerFactorC")
    voltage_a: None | list[None | float] = Field(alias="voltageA")
    voltage_b: None | list[None | float] = Field(alias="voltageB")
    voltage_c: None | list[None | float] = Field(alias="voltageC")

    # Calculate power figures
    power_supply = property(_device_energy_sample_to_power("energy_supplied"))
    power_consumption = property(_device_energy_sample_to_power("energy_consumed"))
    power_consumption_solar = property(
        _device_energy_sample_to_power("energy_consumed_solar")
    )
    power_consumption_battery = property(
        _device_energy_sample_to_power("energy_consumed_battery")
    )

    def __str__(self) -> str:
        """Convert self to a string representation."""
        if self.sample_seconds:
            return (
                f"Import: {self.power_supply:0.02f}kW; "
                f"Export: {self.power_consumption:0.02f}kW ("
                f"Solar: {self.power_consumption_solar:0.02f}kW; "
                f"Battery: {self.power_consumption_battery:0.02f}kW"
                f")"
            )
        else:
            return "Import: _.__kW; Export: _.__kW (Solar: _.__kW; Battery: _.__kW)"


class DeviceReadingsGenericConsumer(DeviceReadings):
    """Readings for a Generic consumer device."""

    device_type: Literal["GENERIC_CONSUMER"] = Field(alias="deviceType")
    operation_status: None | list[None | str] = Field(alias="operationStatus")
    operation_message: None | list[None | str] = Field(alias="operationMessage")

    energy_consumed: None | list[None | float] = Field(alias="energyConsumed")
    energy_consumed_solar: None | list[None | float] = Field(
        alias="energyConsumedSolar"
    )
    energy_consumed_battery: None | list[None | float] = Field(
        alias="energyConsumedBattery"
    )
    energy_consumed_grid: None | list[None | float] = Field(alias="energyConsumedGrid")

    # Calculate power figures
    power_consumption = property(_device_energy_sample_to_power("energy_consumed"))
    power_consumption_solar = property(
        _device_energy_sample_to_power("energy_consumed_solar")
    )
    power_consumption_battery = property(
        _device_energy_sample_to_power("energy_consumed_battery")
    )
    power_consumption_grid = property(
        _device_energy_sample_to_power("energy_consumed_grid")
    )

    def __str__(self) -> str:
        """Convert self to a string representation."""
        if self.sample_seconds:
            return (
                f"Consumption: {self.power_consumption:0.02f}kW ("
                f"Solar: {self.power_consumption_solar:0.02f}kW; "
                f"Battery: {self.power_consumption_battery:0.02f}kW; "
                f"Grid: {self.power_consumption_grid:0.02f}kW"
                f")"
            )
        else:
            return "_.__kW (S: _.__kW; G: _.__kW; B: _.__kW)"


class DeviceReadingsWaterHeater(DeviceReadingsGenericConsumer):
    """Readings for a Water heater device."""

    device_type: Literal["WATER_HEATER"] = Field(alias="deviceType")

    available_energy: None | list[None | float] = Field(alias="availableEnergy")
    max_energy: None | list[None | float] = Field(alias="maxEnergy")
    temp_sensor1: None | list[None | float] = Field(alias="s1")
    temp_sensor2: None | list[None | float] = Field(alias="s2")
    temp_sensor3: None | list[None | float] = Field(alias="s3")
    temp_sensor4: None | list[None | float] = Field(alias="s4")
    temp_sensor5: None | list[None | float] = Field(alias="s5")
    temp_sensor6: None | list[None | float] = Field(alias="s6")
    water_heater_status: None | list[None | dict[str, Any]] = Field(alias="whStatus")

    def __str__(self):
        """Convert instance to string."""
        if self.sample_seconds:
            return (
                f"{super().__str__()}; "
                f"Available: {self.available_energy[-1]}l ({self.energy_ratio:02.0f}%); "
                f"Temp: {self.output_temp:02.01f}℃"
            )
        else:
            return f"{super().__str__()}; Available _l (__%); Temp: __℃"

    @property
    def energy_ratio(self) -> None | float:
        """Ratio of energy available; in %."""
        if self.sample_seconds:
            return round((self.available_energy[-1] / self.max_energy[-1]) * 100, 1)

    @property
    def output_temp(self) -> None | float:
        """Output temperature of water.

        This is the max temp of available sensors.
        """
        if self.sample_seconds:
            return max(
                [
                    self.temp_sensor1[-1],
                    self.temp_sensor2[-1],
                    self.temp_sensor3[-1],
                    self.temp_sensor4[-1],
                    self.temp_sensor5[-1],
                    self.temp_sensor6[-1],
                ]
            )


class DeviceReadingsEnergyBalance(DeviceReadingsGenericConsumer):
    """Readings for the Energy Balance device."""

    device_type: Literal["ENERGY_BALANCE"] = Field(alias="deviceType")


DEVICE_TYPE_MAP = {
    DeviceType.Combiner: DeviceReadingsCombiner,
    DeviceType.SolarPV: DeviceReadingsSolarPV,
    DeviceType.GridMeter: DeviceReadingsGridMeter,
    DeviceType.GenericConsumer: DeviceReadingsGenericConsumer,
    DeviceType.WaterHeater: DeviceReadingsWaterHeater,
    DeviceType.EnergyBalance: DeviceReadingsEnergyBalance,
}


class Readings(BaseModel):
    """Reading history data."""

    range_start: datetime = Field(alias="rangeStart")
    range_end: datetime = Field(alias="rangeEnd")
    range_count: int = Field(alias="rangeCount")
    seconds: int
    installation_id: int = Field(alias="installationId")
    server_time: datetime = Field(alias="serverTime")

    devices: list[DeviceReadings]
    unknown_devices: list[dict[str, Any]]

    def __init__(self, **data):
        """Initialise readings and pre-process devices."""
        raw_devices = data.pop("devices", [])
        data["devices"], data["unknown_devices"] = self._populate_devices(raw_devices)
        super().__init__(**data)

    @staticmethod
    def _populate_devices(
        raw_devices: list[dict[str, Any]]
    ) -> (list[DeviceReadings], list[dict[str, Any]]):
        """Populate known devices and document unknown.

        This is a workaround to handle devices that are not known to the library
        """
        devices = []
        unknown_devices = []
        for raw_device in raw_devices:
            device_type_name = raw_device.get("deviceType")
            if device_type := DEVICE_TYPE_MAP.get(device_type_name):
                try:
                    devices.append(device_type(**raw_device))
                except ValidationError as ex:
                    LOGGER.error("Validation failed: %s", ex)
                    LOGGER.debug("Device data: %s", raw_device)
                    unknown_devices.append(raw_device)
            else:
                LOGGER.warning("Unknown device type: %s", device_type_name)
                LOGGER.debug("Device data: %s", raw_device)
                unknown_devices.append(raw_device)

        return devices, unknown_devices



from datetime import datetime, timezone
from pathlib import Path
from unittest.mock import AsyncMock

import aresponses
from aresponses import ResponsesMockServer
import pytest
from yarl import URL

from combined_energy import client
from combined_energy.models import (
    DeviceReadingsCombiner,
    DeviceReadingsEnergyBalance,
    DeviceReadingsGenericConsumer,
    DeviceReadingsGridMeter,
    DeviceReadingsSolarPV,
    DeviceReadingsWaterHeater,
)


@pytest.fixture
def fixed_time(monkeypatch):
    monkeypatch.setattr(
        client, "now", lambda: datetime(2022, 10, 24, 3, 50, 23, tzinfo=timezone.utc)
    )


def mock_route(
    mock_server: ResponsesMockServer,
    url: URL,
    response: Path = None,
    *,
    method: str = "GET",
    status: int = 200,
    mock_login: bool = True,
    **params: str,
):
    """
    Generate a mock route from a fixture file
    """
    url = URL(url)

    if mock_login:
        mock_server.add(
            "onwatch.combined.energy",
            "/user/Login",
            "POST",
            aresponses.Response(
                text="""{"status": "ok", "jwt": "foo", "expireMins": 30}""",
                status=200,
                content_type="application/json",
            ),
        )
        params["jwt"] = "foo"

    if params:
        url = url.with_query(params)

    mock_server.add(
        url.host,
        url.path_qs,
        method,
        aresponses.Response(
            text=response.read_text(),
            status=status,
            content_type="application/json",
        ),
        match_querystring=True,
    )


class TestCombinedEnergy:
    @pytest.fixture
    def target(self):
        return client.CombinedEnergy("user@example", "pass", installation_id=123)

    @pytest.mark.asyncio
    async def test_user(self, api_responses: Path, aresponses, target):
        mock_route(
            aresponses,
            client.DATA_ACCESS_HOST + "/dataAccess/user",
            api_responses / "user.json",
        )

        actual = await target.user()

        aresponses.assert_plan_strictly_followed()
        assert actual.status == "ok"
        assert actual.user.fullname == "Dave Dobbs"

    @pytest.mark.asyncio
    async def test_installation(self, api_responses: Path, aresponses, target):
        mock_route(
            aresponses,
            client.DATA_ACCESS_HOST + "/dataAccess/installation",
            api_responses / "installation.json",
            i="123",
        )

        actual = await target.installation()

        aresponses.assert_plan_strictly_followed()
        assert actual.status == "ACTIVE"

    @pytest.mark.asyncio
    async def test_installation_customers(
        self, api_responses: Path, aresponses, target
    ):
        mock_route(
            aresponses,
            client.DATA_ACCESS_HOST + "/dataAccess/inst-customers",
            api_responses / "inst-customers.json",
            i="123",
        )

        actual = await target.installation_customers()

        aresponses.assert_plan_strictly_followed()
        assert actual.status == "ok"
        assert actual.customers[0].primary is True

    @pytest.mark.asyncio
    async def test_readings(self, api_responses: Path, aresponses, target):
        mock_route(
            aresponses,
            client.DATA_ACCESS_HOST + "/dataAccess/readings",
            api_responses / "readings.json",
            i="123",
            rangeStart="1666583413",
            rangeEnd="1666583423",
            seconds="5",
        )

        actual = await target.readings(
            datetime(2022, 10, 24, 3, 50, 13, tzinfo=timezone.utc),
            datetime(2022, 10, 24, 3, 50, 23, tzinfo=timezone.utc),
            increment=5,
        )

        aresponses.assert_plan_strictly_followed()
        assert actual.devices[0].device_type == "COMBINER"
        isinstance(actual.devices[0], DeviceReadingsCombiner)
        assert actual.devices[1].device_type == "SOLAR_PV"
        isinstance(actual.devices[1], DeviceReadingsSolarPV)
        assert actual.devices[3].device_type == "GRID_METER"
        isinstance(actual.devices[3], DeviceReadingsGridMeter)
        assert actual.devices[4].device_type == "WATER_HEATER"
        isinstance(actual.devices[4], DeviceReadingsWaterHeater)
        assert actual.devices[5].device_type == "ENERGY_BALANCE"
        isinstance(actual.devices[5], DeviceReadingsEnergyBalance)
        assert actual.devices[6].device_type == "GENERIC_CONSUMER"
        isinstance(actual.devices[6], DeviceReadingsGenericConsumer)

    @pytest.mark.asyncio
    async def test_last_readings(
        self, api_responses: Path, aresponses, target, fixed_time
    ):
        mock_route(
            aresponses,
            client.DATA_ACCESS_HOST + "/dataAccess/readings",
            api_responses / "readings.json",
            i="123",
            rangeStart="1666583123",
            rangeEnd="",
            seconds="5",
        )

        actual = await target.last_readings(
            minutes=5,
            increment=5,
        )

        aresponses.assert_plan_strictly_followed()
        assert actual.range_start == datetime(
            2022, 10, 24, 3, 50, 15, tzinfo=timezone.utc
        )
        assert actual.range_end == datetime(
            2022, 10, 24, 3, 50, 25, tzinfo=timezone.utc
        )

    @pytest.mark.asyncio
    async def test_last_reading__where_no_range_specified(
        self, api_responses: Path, target
    ):
        with pytest.raises(ValueError):
            await target.last_readings(increment=5)

    @pytest.mark.asyncio
    async def test_communication_status(self, api_responses: Path, aresponses, target):
        mock_route(
            aresponses,
            client.DATA_ACCESS_HOST + "/dataAccess/comm-stat",
            api_responses / "comm-stat.json",
            i="123",
        )

        actual = await target.communication_status()

        aresponses.assert_plan_strictly_followed()
        assert actual.status == "ok"
        assert actual.connected is True
        assert actual.since == datetime(2022, 10, 24, 3, 50, 13, tzinfo=timezone.utc)

    @pytest.mark.asyncio
    async def test_communication_history(self, api_responses: Path, aresponses, target):
        aresponses.add(
            "onwatch.combined.energy",
            "/user/Login",
            "POST",
            aresponses.Response(
                text="""{"status": "ok", "jwt": "foo", "expireMins": 30}""",
                status=200,
                content_type="application/json",
            ),
        )
        aresponses.add(
            "ds20.combined.energy",
            "/data-service/dataAccess/comm-hist",
            "GET",
            aresponses.Response(
                text=(api_responses / "comm-hist.json").read_text(),
                status=200,
                content_type="application/json",
            ),
        )

        actual = await target.communication_history()

        aresponses.assert_plan_strictly_followed()
        assert actual.status == "ok"
        assert actual.history[0].timestamp == datetime(
            2022, 10, 24, 3, 50, 13, 254000, tzinfo=timezone.utc
        )

    @pytest.mark.asyncio
    async def test_login__where_login_is_successful(
        self, api_responses: Path, aresponses, target
    ):
        aresponses.add(
            "onwatch.combined.energy",
            "/user/Login",
            "POST",
            aresponses.Response(
                text="""{"status": "ok", "jwt": "foo", "expireMins": 30}""",
                status=200,
                content_type="application/json",
            ),
        )

        actual = await target.login()

        aresponses.assert_plan_strictly_followed()
        assert actual.jwt == "foo"

    @pytest.mark.asyncio
    async def test_login__where_login_fails(self, target, aresponses):
        aresponses.add(
            "onwatch.combined.energy",
            "/user/Login",
            "POST",
            aresponses.Response(
                text="""{"status": "failed", "error": "bad password"}""",
                status=200,
                content_type="application/json",
            ),
        )

        with pytest.raises(client.exceptions.CombinedEnergyAuthError):
            await target.login()

        aresponses.assert_plan_strictly_followed()

    @pytest.mark.asyncio
    async def test_start_log_session__where_refresh_is_successful(
        self, api_responses, target, aresponses
    ):
        aresponses.add(
            "onwatch.combined.energy",
            "/user/Login",
            "POST",
            aresponses.Response(
                text="""{"status": "ok", "jwt": "foo", "expireMins": 30}""",
                status=200,
                content_type="application/json",
            ),
        )
        aresponses.add(
            "dp20.combined.energy",
            "/mqtt2/user/LogSessionStart",
            "POST",
            aresponses.Response(
                text=(api_responses / "LogSessionStart.json").read_text(),
                status=200,
                content_type="application/json",
            ),
        )

        await target.start_log_session()

        aresponses.assert_plan_strictly_followed()

    @pytest.mark.asyncio
    async def test_request__with_permission_denied(
        self, target, aresponses, api_responses
    ):
        mock_route(
            aresponses,
            client.DATA_ACCESS_HOST + "/dataAccess/comm-stat",
            api_responses / "comm-stat.json",
            status=401,
            i="123",
        )

        with pytest.raises(client.exceptions.CombinedEnergyPermissionError):
            await target.communication_status()

        aresponses.assert_plan_strictly_followed()

    @pytest.mark.asyncio
    async def test_request__with_server_error(
        self, target, aresponses, api_responses, caplog
    ):
        caplog.at_level("ERROR", client.LOGGER)
        mock_route(
            aresponses,
            client.DATA_ACCESS_HOST + "/dataAccess/comm-stat",
            api_responses / "comm-stat.json",
            status=500,
            i="123",
        )

        with pytest.raises(client.exceptions.CombinedEnergyError):
            await target.communication_status()

        aresponses.assert_plan_strictly_followed()
        assert len(caplog.messages) == 1
        assert caplog.messages[0].startswith("Server error")

    @pytest.mark.asyncio
    async def test_request__with_client_error(
        self, target, aresponses, api_responses, caplog
    ):
        caplog.at_level("ERROR", client.LOGGER)
        mock_route(
            aresponses,
            client.DATA_ACCESS_HOST + "/dataAccess/comm-stat",
            api_responses / "comm-stat.json",
            i="123",
            eek="!",
        )

        with pytest.raises(client.exceptions.CombinedEnergyError):
            await target.communication_status()

        assert len(caplog.messages) == 2
        assert caplog.messages[1].startswith("Socket error")

    @pytest.mark.asyncio
    async def test_async_context_manager__external_session(self):
        async with client.CombinedEnergy(
            "user@example", "pass", installation_id=123, session=AsyncMock()
        ) as target:
            assert target.session is not None

        assert target.session is not None

    @pytest.mark.asyncio
    async def test_async_context_manager__internal_session(self):
        session = AsyncMock()

        async with client.CombinedEnergy(
            "user@example", "pass", installation_id=123
        ) as target:
            target.session = session
            target._close_session = True

        assert target.session is None
        session.close.assert_called()



"""Constants used by API."""
from enum import Enum
import logging

LOGGER = logging.getLogger(__package__)


class DeviceType(str, Enum):
    """Type of device."""

    Battery = "BATTERY"
    Combiner = "COMBINER"
    EnergyPredicted = "ENERGY_PRED"
    Receiver = "DRED_RECEIVER"  # Not sure what this is
    EnergyBalance = "ENERGY_BALANCE"
    GenericConsumer = "GENERIC_CONSUMER"
    GridMeter = "GRID_METER"
    Monitor = "MONITOR"  # Not sure what this is
    PoolHeater = "POOL_HEATER"
    SolarPredicted = "SOLAR_PRED"
    SolarPV = "SOLAR_PV"
    Tank = "TANKPAK"  # Not sure what this is
    Total = "TOTAL"
    WaterHeater = "WATER_HEATER"


class Category(str, Enum):
    """Category of device."""

    Airconditioner = "AIRCON"
    Battery = "BATTERY"
    Building = "BUILDING"
    CarCharger = "CAR_CHARGER"
    Combiner = "COMBINER"
    Cooking = "COOKING"
    GridMeter = "GRID_METER"
    Heating = "HEATING"
    Monitor = "MONITOR"
    Others = "OTHERS"
    Pool = "POOL"
    SolarPV = "SOLAR_PV"
    Tank = "TANKPAK"
    WaterHeater = "WATER_HEATER"



"""Client for API."""
from __future__ import annotations

import asyncio
from dataclasses import dataclass
import datetime
from importlib import metadata
import logging
import socket

try:
    # Support builtin asyncio timeout in Python 3.9+
    from asyncio import timeout as aio_timeout
except ImportError:
    from async_timeout import timeout as aio_timeout

from aiohttp import ClientError, ClientResponseError, ClientSession
from aiohttp.hdrs import METH_GET, METH_POST

from . import exceptions
from .constants import LOGGER
from .models import (
    ConnectionHistory,
    ConnectionStatus,
    CurrentUser,
    Installation,
    InstallationCustomers,
    Login,
    Readings,
)

USER_ACCESS_HOST = "https://onwatch.combined.energy"
DATA_ACCESS_HOST = "https://ds20.combined.energy/data-service"
MQTT_ACCESS_HOST = "https://dp20.combined.energy"
now = datetime.datetime.now


@dataclass
class CombinedEnergy:
    """Client library for Combined Energy API."""

    mobile_or_email: str
    password: str
    installation_id: int

    # Configuration Options
    expiry_window: int = 300
    request_timeout: float = 10.0

    # State
    session: ClientSession | None = None
    _close_session: bool = False
    _jwt: str | None = None
    _expires: datetime | None = None

    async def _make_request(
        self,
        url: str,
        params: dict[str, str] = None,
        data: dict[str, str] = None,
        *,
        method: str = METH_GET,
        accept: str = "application/json",
        request_timeout: float = None,
    ):
        """Handle a request to the Combined Energy API."""
        version = metadata.version("combined-energy-api")
        headers = {
            "Accept": accept,
            "User-Agent": f"PythonCombinedEnergy/{version}",
        }

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug(
                "Request to %s?%s",
                url,
                "&".join(f"{k}={v}" for k, v in (params or {}).items()),
            )

        try:
            async with aio_timeout(request_timeout or self.request_timeout):
                response = await self.session.request(
                    method, url, params=params, headers=headers, data=data
                )
                if response.status == 500:
                    # Trigger response.text to ensure error is captured
                    LOGGER.error("Server error: %s", await response.text())
                response.raise_for_status()

        except asyncio.TimeoutError as ex:
            raise exceptions.CombinedEnergyTimeoutError(
                "Timeout occurred while connecting to the Combined Energy API"
            ) from ex

        except ClientResponseError as ex:
            if ex.status == 401:
                raise exceptions.CombinedEnergyPermissionError(
                    "Current user does not have access to the specified resource"
                ) from ex

            raise exceptions.CombinedEnergyError(
                f"Unexpected error: {ex.status}", ex.status
            ) from ex

        except (
            ClientError,
            socket.gaierror,
        ) as ex:
            LOGGER.error("Socket error: %s", ex)
            raise exceptions.CombinedEnergyError(
                "Error occurred while communicating with the Combined Energy API"
            ) from ex

        return await response.json()

    async def _get_token(self):
        """Get JWT."""
        login = await self.login()
        self._jwt = login.jwt
        self._expires = login.expires(self.expiry_window)

    async def _ensure_token(self):
        """Check if token is required."""
        if not self._jwt or now() > self._expires:
            LOGGER.info("Login expired; re-login" if self._jwt else "Login required")
            await self._get_token()

    async def _request(self, url: str, **params):
        # Apply token
        await self._ensure_token()
        params.setdefault("jwt", self._jwt)

        return await self._make_request(url, params)

    async def user(
        self,
    ) -> CurrentUser:
        """Get details of current user."""
        data = await self._request(
            DATA_ACCESS_HOST + "/dataAccess/user",
        )
        return CurrentUser.model_validate(data)

    async def installation(
        self,
    ) -> Installation:
        """Get details of installation."""
        data = await self._request(
            DATA_ACCESS_HOST + "/dataAccess/installation",
            i=self.installation_id,
        )
        return Installation.model_validate(data)

    async def installation_customers(
        self,
    ) -> InstallationCustomers:
        """Get list of customers associated with an installation."""
        data = await self._request(
            DATA_ACCESS_HOST + "/dataAccess/inst-customers",
            i=self.installation_id,
        )
        return InstallationCustomers.model_validate(data)

    async def readings(
        self,
        range_start: datetime.datetime | None,
        range_end: datetime.datetime | None,
        increment: int,
    ) -> Readings:
        """Get readings from system.

        :param range_start: Start of readings range
        :param range_end: End of readings range
        :param increment: Sample increment size in seconds
        """
        data = await self._request(
            DATA_ACCESS_HOST + "/dataAccess/readings",
            i=self.installation_id,
            rangeStart=int(range_start.timestamp()) if range_start else "",
            rangeEnd=int(range_end.timestamp()) if range_end else "",
            seconds=increment,
        )
        readings = Readings.model_validate(data)

        return readings

    async def last_readings(
        self,
        hours: float = 0,
        minutes: float = 0,
        seconds: float = 0,
        *,
        increment: int,
    ):
        """Get readings for the last delta range.

        :param hours: Delta in hours
        :param minutes: Delta in minutes
        :param seconds: Delta in seconds
        :param increment: Sample increment size in seconds
        """

        delta = datetime.timedelta(hours=hours, minutes=minutes, seconds=seconds)
        if not delta.total_seconds():
            raise ValueError("Either a time range must be provided")
        return await self.readings(now() - delta, None, increment)

    # getPerformanceSummary: dataHost + "/dataAccess/performance-summary",
    # getTariffDetails: dataHost + "/dataAccess/tariff-details",

    async def communication_status(self) -> ConnectionStatus:
        """Get communication status of the installation."""
        data = await self._request(
            DATA_ACCESS_HOST + "/dataAccess/comm-stat",
            i=self.installation_id,
        )
        return ConnectionStatus.model_validate(data)

    async def communication_history(self) -> ConnectionHistory:
        """Get communication history of the installation."""
        data = await self._request(
            DATA_ACCESS_HOST + "/dataAccess/comm-hist",
            i=self.installation_id,
        )
        return ConnectionHistory.model_validate(data)

    async def login(self) -> Login:
        """Login and obtain a web token."""
        data = await self._make_request(
            USER_ACCESS_HOST + "/user/Login",
            data={
                "mobileOrEmail": self.mobile_or_email,
                "pass": self.password,
                "store": False,
            },
            method=METH_POST,
        )
        if data.get("status") != "ok":
            raise exceptions.CombinedEnergyAuthError(data.get("error", "Login failed"))
        return Login.model_validate(data)

    async def start_log_session(self) -> bool:
        """Trigger the start of a log session (required if readings stop being supplied)."""
        await self._ensure_token()

        try:
            data = await self._make_request(
                MQTT_ACCESS_HOST + "/mqtt2/user/LogSessionStart",
                data={
                    "i": self.installation_id,
                    "jwt": self._jwt,
                },
                method=METH_POST,
            )
            return data.get("status") == "ok"

        except exceptions.CombinedEnergyTimeoutError as ex:
            # Not sure why this particular request consistently times out, ignoring
            # the response brings it back to life
            LOGGER.warning("LogSessionStart request timed out with: %s", ex)
            return True  # Assume ok

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()
            self.session = None

    async def __aenter__(self) -> CombinedEnergy:
        """Async context manager aenter implementation."""
        return self

    async def __aexit__(self, *_exc_info):
        """Async context manager aexit implementation."""
        await self.close()



from datetime import datetime
from unittest.mock import Mock

from combined_energy import models
import pytest


class TestLogin:
    def test_expires(self):
        target = models.Login(
            status="ok",
            expireMins=30,
            jwt="...",
            created=datetime(2022, 10, 26, 1, 44),
        )

        actual = target.expires(60)

        assert actual == datetime(2022, 10, 26, 2, 13)


def test_device_energy_sample_to_power__where_reading_has_data():
    mock_device_readings = Mock(
        models.DeviceReadings, energy_readings=[2.3, 5.0], sample_seconds=[5, 5]
    )

    target = models._device_energy_sample_to_power("energy_readings")

    actual = target(mock_device_readings)

    assert actual == 3.6


def test_device_energy_sample_to_power__where_no_data():
    mock_device_readings = Mock(
        models.DeviceReadings, energy_readings=None, sample_seconds=[]
    )

    target = models._device_energy_sample_to_power("energy_readings")

    actual = target(mock_device_readings)

    assert actual is None


class TestDeviceReadingsWaterHeater:
    @pytest.fixture
    def target_available(self) -> models.DeviceReadingsWaterHeater:
        return models.DeviceReadingsWaterHeater(
            deviceId=1,
            rangeStart=None,
            rangeEnd=None,
            timestamp=[datetime.now(), datetime.now()],
            sampleSecs=[5, 5],
            operationStatus=[None, None],
            operationMessage=[None, None],
            energyConsumed=[1, 1],
            energyConsumedSolar=[1, 1],
            energyConsumedBattery=[1, 1],
            energyConsumedGrid=[1, 1],
            deviceType="WATER_HEATER",
            availableEnergy=[300, 491.4],
            maxEnergy=[630, 630],
            s1=[16.5, 17.5],
            s2=[16.5, 23.5],
            s3=[16.5, 50.8],
            s4=[16.5, 55.8],
            s5=[16.5, 60.3],
            s6=[16.5, 68.9],
            whStatus=[None, None],
        )

    @pytest.fixture
    def target_not_available(self) -> models.DeviceReadingsWaterHeater:
        return models.DeviceReadingsWaterHeater(
            deviceId=1,
            rangeStart=None,
            rangeEnd=None,
            timestamp=[],
            sampleSecs=[],
            operationStatus=None,
            operationMessage=None,
            energyConsumed=None,
            energyConsumedSolar=None,
            energyConsumedBattery=None,
            energyConsumedGrid=None,
            deviceType="WATER_HEATER",
            availableEnergy=None,
            maxEnergy=None,
            s1=None,
            s2=None,
            s3=None,
            s4=None,
            s5=None,
            s6=None,
            whStatus=None,
        )

    def test_energy_ratio__where_data_available(self, target_available):
        actual = target_available.energy_ratio

        assert actual == 78.0

    def test_energy_ratio__where_no_data(self, target_not_available):
        actual = target_not_available.energy_ratio

        assert actual is None

    def test_output_temp__where_data_available(self, target_available):
        actual = target_available.output_temp

        assert actual == 68.9

    def test_output_temp__where_no_data(self, target_not_available):
        actual = target_not_available.output_temp

        assert actual is None



"""Utility methods."""
from typing import Final

KWH_TO_KW: Final[float] = 3.6


def energy_to_power(energy: float, seconds: float) -> float:
    """
    Convert an energy value to power.

    Essentially take an energy value in kWh and calculate an instant power in kW.
    """
    if seconds <= 0.0:
        raise ValueError("Seconds must be a positive float.")

    return energy * KWH_TO_KW / seconds
